==> atmsupport-1.0.tar.gz: atmsupport-1.0/test_uni_user/main.c <==
/*
 * Copyright (c) 2001-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Author: Hartmut Brandt <harti@freebsd.org>
 *
 * $Begemot: atmsupport/test_uni_user/main.c,v 1.3 2004/01/02 15:15:56 hbb Exp $
 *
 * Test user space UNI stack.
 */
#include "uniu.h"

#define MAXMSG		8192

struct timer {
	evTimerID	id;
	struct stk	*stk;
	void		(*func)(void *);
	void		*arg;
};

static int	debug = 1;
static const char *script = "uniu.tcl";
static const char *tclpath = ".";
static Tcl_DString command;
int errinfo = 1;
int doprompt = 1;

/*
 * Input from ATM socket.
 */
static void
atm_input(evContext ctx __unused, void *uap, int fd, int eventmask __unused)
{
	struct stk *stk = (struct stk *)uap;
	struct uni_msg *m;
	ssize_t n;

	if ((m = uni_msg_alloc(MAXMSG + 1)) == NULL)
		err(1, "cannot allocate input message");
	if ((n = read(fd, m->b_rptr, MAXMSG + 1)) == -1)
		err(1, "input read error");
	if ((size_t)n == MAXMSG + 1)
		warnx("input message too large");

	if (debug)
		fprintf(stderr, "ATM input %zu bytes\n", (size_t)n);

	m->b_wptr = m->b_rptr + (size_t)n;
	sscop_input(stk->sscop, m);
}

/*
 * Output to ATM socket
 */
static void
sscop_send_lower(struct sscop *sscop __unused, void *arg, struct uni_msg *m)
{
	struct stk *stk = (struct stk *)arg;
	ssize_t n;

	if ((n = write(stk->atm_fd, m->b_rptr, uni_msg_len(m))) == -1)
		err(1, "atm write error");
	if ((size_t)n != uni_msg_len(m))
		errx(1, "short write %zu %zu", (size_t)n, uni_msg_len(m));

	if (debug)
		fprintf(stderr, "ATM output %zu bytes\n", (size_t)n);
	uni_msg_destroy(m);
}

/*
 * SSCOP management signal
 */
static void
sscop_send_manage(struct sscop *sscop __unused, void *arg __unused,
    enum sscop_maasig sig, struct uni_msg *m, u_int code, u_int cnt)
{
	if (debug)
		fprintf(stderr, "SSCOP MAA SIG %s\n", sscop_msigname(sig));

	switch (sig) {

	  case SSCOP_MERROR_indication:
		warnx("SSCOP error %c,%u\n", (char)code, cnt);
		break;

	  case SSCOP_MDATA_indication:
		fprintf(stderr, "SSCOP management data %zu bytes\n",
		    uni_msg_len(m));
		uni_msg_destroy(m);
		break;

	  case SSCOP_MDATA_request:
		abort();
	}
}

/*
 * SSCOP signal - relay to SSCF
 */
static void
sscop_send_upper(struct sscop *sscop __unused, void *arg,
    enum sscop_aasig sig, struct uni_msg *m, u_int n)
{
	struct stk *stk = (struct stk *)arg;

	if (debug)
		fprintf(stderr, "SSCOP AA SIG %s\n", sscop_signame(sig));

	sscfu_input(stk->sscf, sig, m, n);
}

/*
 * sscop timer helper
 */
static void
sscop_tfunc(evContext ctx __unused, void *uap, struct timespec due __unused,
    struct timespec inter __unused)
{
	struct timer *t = uap;

	t->func(t->stk->sscop);
	free(t);
}

/*
 * Start an sscop timer
 */
static void *
sscop_start_timer(struct sscop *sscop __unused, void *arg, u_int msec,
    void (*func)(void *))
{
	struct stk *stk = (struct stk *)arg;
	struct timer *t;
	struct timespec due;

	if ((t = malloc(sizeof(*t))) == NULL)
		err(1, "out of memory");
	t->stk = stk;
	t->func = func;

	due = evAddTime(evNowTime(),
	    evConsTime((time_t)msec/1000, (long)(msec%1000)*1000));

	if (evSetTimer(stk->ctx, sscop_tfunc, t, due, evConsTime(0, 0), &t->id))
		err(1, "cannot start timer");

	return (t);
}

/*
 * Stop an sscop timer
 */
static void
sscop_stop_timer(struct sscop *sscop __unused, void *arg, void *tp)
{
	struct stk *stk = (struct stk *)arg;
	struct timer *t = tp;

	evClearTimer(stk->ctx, t->id);
	free(t);
}

/*
 * sscop timer helper
 */
static void
uni_tfunc(evContext ctx __unused, void *uap, struct timespec due __unused,
    struct timespec inter __unused)
{
	struct timer *t = uap;

	t->func(t->arg);
	free(t);
}

/*
 * Start an uni timer
 */
static void *
uni_start_timer(struct uni *uni __unused, void *arg, u_int msec,
    void (*func)(void *), void *farg)
{
	struct stk *stk = (struct stk *)arg;
	struct timer *t;
	struct timespec due;

	if ((t = malloc(sizeof(*t))) == NULL)
		err(1, "out of memory");
	t->stk = stk;
	t->func = func;
	t->arg = farg;

	due = evAddTime(evNowTime(),
	    evConsTime((time_t)msec/1000, (long)(msec%1000)*1000));

	if (evSetTimer(stk->ctx, uni_tfunc, t, due, evConsTime(0, 0), &t->id))
		err(1, "cannot start timer");

	return (t);
}

/*
 * Stop an uni timer
 */
static void
uni_stop_timer(struct uni *uni __unused, void *arg, void *tp)
{
	struct stk *stk = (struct stk *)arg;
	struct timer *t = tp;

	evClearTimer(stk->ctx, t->id);
	free(t);
}

/*
 * Signal from SSCF to lower layer - relay to SSCOP
 */
static void
sscfu_send_lower(struct sscfu *sscf __unused, void *arg, enum sscop_aasig sig,
    struct uni_msg *m, u_int n)
{
	struct stk *stk = (struct stk *)arg;

	if (debug)
		fprintf(stderr, "SSCOP AA SIG %s\n", sscop_signame(sig));

	sscop_aasig(stk->sscop, sig, m, n);
}

/*
 * Signal from SSCF to upper layer - relay to UNI
 */
static void
sscfu_send_upper(struct sscfu *sscf __unused, void *arg, enum saal_sig sig,
    struct uni_msg *m)
{
	struct stk *stk = (struct stk *)arg;

	if (debug)
		fprintf(stderr, "SAAL SIG %s\n", sscfu_signame(sig));

	uni_saal_input(stk->uni, sig, m);
	uni_work(stk->uni);
}

/*
 * Move SSCOP window - relay to SSCOP
 */
static void
sscfu_window(struct sscfu *sscf __unused, void *arg, u_int w)
{
	struct stk *stk = (struct stk *)arg;

	(void)sscop_window(stk->sscop, w);
}

/*
 * Output from the UNI to the SAAL - relay to SSCF
 */
static void
uni_saal_output(struct uni *uni __unused, void *arg, enum saal_sig sig,
    struct uni_msg *m)
{
	struct stk *stk = (struct stk *)arg;

	(void)sscfu_saalsig(stk->sscf, sig, m);
}

/*
 * Output from the UNI to the upper layer - relay to user
 */
static void
uni_uni_output(struct uni *uni __unused, void *arg, enum uni_sig sig,
    u_int32_t cookie, struct uni_msg *m)
{
	printf("UNI sig %s\n", uni_signame(sig));
	user_output(arg, sig, cookie, m);
	if (m != NULL)
		uni_msg_destroy(m);
}

/*
 * Input from user
 */
static void
user_input(evContext ctx __unused, void *uap, int fd, int eventmask __unused)
{
	struct stk *stk = (struct stk *)uap;
	char line[LINE_MAX], *ptr, *end;
	ssize_t n;

	if ((n = read(fd, line, sizeof(line))) == -1)
		err(1, "read from user");
	if (n == 0) {
		Tcl_DStringFree(&command);
		Tcl_DStringAppend(&command, "exit", 4);
		try_exec(stk->interp, &command);
		exit(2);
	}
	ptr = line;
	while (n >= 0) {
		for (end = ptr; end < ptr + n && *end != '\n'; end++)
			;
		if (end == ptr + n) {
			Tcl_DStringAppend(&command, ptr, n);
			break;
		}
		Tcl_DStringAppend(&command, ptr, end - ptr);
		try_exec(stk->interp, &command);
		n -= end - ptr + 1;
		ptr = end + 1;
	}
}

static void
sscop_verbose(struct sscop *sscop __unused, void *arg __unused,
    const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	fprintf(stderr, "sscop: ");
	vfprintf(stderr, fmt, ap);
	fprintf(stderr, "\n");
	va_end(ap);
}

static void
sscfu_verbose(struct sscfu *sscfu __unused, void *arg __unused,
    const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	fprintf(stderr, "sscf: ");
	vfprintf(stderr, fmt, ap);
	fprintf(stderr, "\n");
	va_end(ap);
}

static void
uni_verbose(struct uni *uni __unused, void *arg __unused,
    enum uni_verb fac, const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	fprintf(stderr, "uni(%s): ", uni_facname(fac));
	vfprintf(stderr, fmt, ap);
	fprintf(stderr, "\n");
	va_end(ap);
}

static void
uni_status_func(struct uni *uni __unused, void *uarg __unused,
    void *arg __unused, const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
}


static const struct sscop_funcs sscop_funcs = {
	sscop_send_manage,
	sscop_send_upper,
	sscop_send_lower,
	sscop_verbose,
	sscop_start_timer,
	sscop_stop_timer
};

static const struct sscfu_funcs sscfu_funcs = {
	sscfu_send_upper,
	sscfu_send_lower,
	sscfu_window,
	sscfu_verbose
};

static const struct uni_funcs uni_funcs = {
	uni_uni_output,
	uni_saal_output,
	uni_verbose,
	uni_status_func,
	uni_start_timer,
	uni_stop_timer
};

int
main(int argc __unused, char *argv[] __unused)
{
	struct stk stk;
	evEvent ev;
	Tcl_DString args;
	char *str;
	const char *res;
	int opt;

	memset(&stk, 0, sizeof(stk));
	stk.atm_fd = STDIN_FILENO;
	stk.user_fd = STDOUT_FILENO;

	argc--;
	argv++;

	if ((stk.sscop = sscop_create(&stk, &sscop_funcs)) == NULL)
		err(1, "cannot create sscop instance");
	if ((stk.sscf = sscfu_create(&stk, &sscfu_funcs)) == NULL)
		err(1, "cannot create sscf instance");
	if ((stk.uni = uni_create(&stk, &uni_funcs)) == NULL)
		err(1, "cannot create uni instance");

	if (evCreate(&stk.ctx))
		err(1, "evCreate");

	if ((stk.interp = Tcl_CreateInterp()) == NULL)
		err(1, "cannot create TCL interpreter");
	Tcl_DStringInit(&command);

	user_tcl_init(&stk);

	/*
	 * Make script name available as 'script'
	 */
	if((str = malloc(strlen(script)+1)) == NULL)
		err(1, NULL);
	script = strcpy(str, script);
	if(Tcl_LinkVar(stk.interp, "script", (char *)&script, 
	    TCL_LINK_STRING|TCL_LINK_READ_ONLY) != TCL_OK)
		errx(1, "Tcl_LinkVar(script): %s",
		    Tcl_GetStringResult(stk.interp));

	/*
	 * Make the execution path of the script availablle as 'tclpath'
	 */
	if((str = malloc(strlen(tclpath)+1)) == NULL)
		err(1, NULL);
	tclpath = strcpy(str, tclpath);
	if(Tcl_LinkVar(stk.interp, "tclpath", (char *)&tclpath, 
	    TCL_LINK_STRING|TCL_LINK_READ_ONLY) != TCL_OK)
		errx(1, "Tcl_LinkVar(tclpath): %s",
		    Tcl_GetStringResult(stk.interp));

	/*
	 * Make the program arguments available
	 */
	Tcl_DStringInit(&args);
	while(argc--)
		Tcl_DStringAppendElement(&args, *argv++);
	Tcl_SetVar(stk.interp, "argv", Tcl_DStringValue(&args), 0);
	Tcl_DStringFree(&args);

	/*
	 * Run startup script
	 */
	asprintf(&str, "%s/%s", tclpath, script);
	if (str == NULL)
		err(1, "asprintf");
	opt = Tcl_EvalFile(stk.interp, str);
	free(str);

	if (*(res = Tcl_GetStringResult(stk.interp)) != 0)
		fprintf(stderr, "%s\n", res);
	if (opt != TCL_OK) {
		fprintf(stderr, "initialisation returned %d\n", opt);
		if (errinfo) {
			if ((res = Tcl_GetVar(stk.interp, "errorInfo",
			    TCL_GLOBAL_ONLY)) != NULL)
				fprintf(stderr, "errorInfo='%s'\n", res);
			if ((res = Tcl_GetVar(stk.interp, "errorCode",
			    TCL_GLOBAL_ONLY)) != NULL)
				fprintf(stderr, "errorCode='%s'\n", res);
		}
		exit(1);
	}

	/*
	 * Start file descriptors
	 */
	if (evSelectFD(stk.ctx, stk.atm_fd, EV_READ, atm_input, &stk, &stk.atm_fdh))
		err(1, "select ATM file");

	if(evSelectFD(stk.ctx, stk.user_fd, EV_READ, user_input, &stk, &stk.user_fdh))
		err(1, "select USER file");

	doprompt = 1;
	while (1) {
		if (doprompt && evTestID(stk.user_fdh) && isatty(stk.user_fd)) {
			write(stk.user_fd, "@> ", 3);
			doprompt = 0;
		}

		if (evGetNext(stk.ctx, &ev, EV_WAIT) == 0) {
			if (evDispatch(stk.ctx, ev))
				err(1, "dispatch event");
		} else if (errno != EINTR)
			err(1, "get event");
	}

	return (0);
}

==> atmsupport-1.4.tar.gz: atmsupport-1.4/test_uni_user/main.c <==
/*
 * Copyright (c) 2001-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Author: Hartmut Brandt <harti@freebsd.org>
 *
 * $Begemot: atmsupport/test_uni_user/main.c,v 1.4 2004/12/20 10:45:06 brandt Exp $
 *
 * Test user space UNI stack.
 */
#include "uniu.h"

#define MAXMSG		8192

struct timer {
	int		id;
	struct stk	*stk;
	void		(*func)(void *);
	void		*arg;
};

static int	debug = 1;
static const char *script = "uniu.tcl";
static const char *tclpath = ".";
static Tcl_DString command;
int errinfo = 1;
int doprompt = 1;

/*
 * Input from ATM socket.
 */
static void
atm_input(int fd, int eventmask __unused, void *uap)
{
	struct stk *stk = (struct stk *)uap;
	struct uni_msg *m;
	ssize_t n;

	if ((m = uni_msg_alloc(MAXMSG + 1)) == NULL)
		err(1, "cannot allocate input message");
	if ((n = read(fd, m->b_rptr, MAXMSG + 1)) == -1)
		err(1, "input read error");
	if ((size_t)n == MAXMSG + 1)
		warnx("input message too large");

	if (debug)
		fprintf(stderr, "ATM input %zu bytes\n", (size_t)n);

	m->b_wptr = m->b_rptr + (size_t)n;
	sscop_input(stk->sscop, m);
}

/*
 * Output to ATM socket
 */
static void
sscop_send_lower(struct sscop *sscop __unused, void *arg, struct uni_msg *m)
{
	struct stk *stk = (struct stk *)arg;
	ssize_t n;

	if ((n = write(stk->atm_fd, m->b_rptr, uni_msg_len(m))) == -1)
		err(1, "atm write error");
	if ((size_t)n != uni_msg_len(m))
		errx(1, "short write %zu %zu", (size_t)n, uni_msg_len(m));

	if (debug)
		fprintf(stderr, "ATM output %zu bytes\n", (size_t)n);
	uni_msg_destroy(m);
}

/*
 * SSCOP management signal
 */
static void
sscop_send_manage(struct sscop *sscop __unused, void *arg __unused,
    enum sscop_maasig sig, struct uni_msg *m, u_int code, u_int cnt)
{
	if (debug)
		fprintf(stderr, "SSCOP MAA SIG %s\n", sscop_msigname(sig));

	switch (sig) {

	  case SSCOP_MERROR_indication:
		warnx("SSCOP error %c,%u\n", (char)code, cnt);
		break;

	  case SSCOP_MDATA_indication:
		fprintf(stderr, "SSCOP management data %zu bytes\n",
		    uni_msg_len(m));
		uni_msg_destroy(m);
		break;

	  case SSCOP_MDATA_request:
		abort();
	}
}

/*
 * SSCOP signal - relay to SSCF
 */
static void
sscop_send_upper(struct sscop *sscop __unused, void *arg,
    enum sscop_aasig sig, struct uni_msg *m, u_int n)
{
	struct stk *stk = (struct stk *)arg;

	if (debug)
		fprintf(stderr, "SSCOP AA SIG %s\n", sscop_signame(sig));

	sscfu_input(stk->sscf, sig, m, n);
}

/*
 * sscop timer helper
 */
static void
sscop_tfunc(int tid __unused, void *uap)
{
	struct timer *t = uap;

	t->func(t->stk->sscop);
	free(t);
}

/*
 * Start an sscop timer
 */
static void *
sscop_start_timer(struct sscop *sscop __unused, void *arg, u_int msec,
    void (*func)(void *))
{
	struct stk *stk = (struct stk *)arg;
	struct timer *t;

	if ((t = malloc(sizeof(*t))) == NULL)
		err(1, "out of memory");
	t->stk = stk;
	t->func = func;
	t->id = poll_start_timer(msec, 0, sscop_tfunc, t);

	return (t);
}

/*
 * Stop an sscop timer
 */
static void
sscop_stop_timer(struct sscop *sscop __unused, void *arg __unused, void *tp)
{
	struct timer *t = tp;

	poll_stop_timer(t->id);
	free(t);
}

/*
 * sscop timer helper
 */
static void
uni_tfunc(int tid __unused, void *uap)
{
	struct timer *t = uap;

	t->func(t->arg);
	free(t);
}

/*
 * Start an uni timer
 */
static void *
uni_start_timer(struct uni *uni __unused, void *arg, u_int msec,
    void (*func)(void *), void *farg)
{
	struct stk *stk = (struct stk *)arg;
	struct timer *t;

	if ((t = malloc(sizeof(*t))) == NULL)
		err(1, "out of memory");
	t->stk = stk;
	t->func = func;
	t->arg = farg;
	t->id = poll_start_timer(msec, 0, uni_tfunc, t);

	return (t);
}

/*
 * Stop an uni timer
 */
static void
uni_stop_timer(struct uni *uni __unused, void *arg __unused, void *tp)
{
	struct timer *t = tp;

	poll_stop_timer(t->id);
	free(t);
}

/*
 * Signal from SSCF to lower layer - relay to SSCOP
 */
static void
sscfu_send_lower(struct sscfu *sscf __unused, void *arg, enum sscop_aasig sig,
    struct uni_msg *m, u_int n)
{
	struct stk *stk = (struct stk *)arg;

	if (debug)
		fprintf(stderr, "SSCOP AA SIG %s\n", sscop_signame(sig));

	sscop_aasig(stk->sscop, sig, m, n);
}

/*
 * Signal from SSCF to upper layer - relay to UNI
 */
static void
sscfu_send_upper(struct sscfu *sscf __unused, void *arg, enum saal_sig sig,
    struct uni_msg *m)
{
	struct stk *stk = (struct stk *)arg;

	if (debug)
		fprintf(stderr, "SAAL SIG %s\n", sscfu_signame(sig));

	uni_saal_input(stk->uni, sig, m);
	uni_work(stk->uni);
}

/*
 * Move SSCOP window - relay to SSCOP
 */
static void
sscfu_window(struct sscfu *sscf __unused, void *arg, u_int w)
{
	struct stk *stk = (struct stk *)arg;

	(void)sscop_window(stk->sscop, w);
}

/*
 * Output from the UNI to the SAAL - relay to SSCF
 */
static void
uni_saal_output(struct uni *uni __unused, void *arg, enum saal_sig sig,
    struct uni_msg *m)
{
	struct stk *stk = (struct stk *)arg;

	(void)sscfu_saalsig(stk->sscf, sig, m);
}

/*
 * Output from the UNI to the upper layer - relay to user
 */
static void
uni_uni_output(struct uni *uni __unused, void *arg, enum uni_sig sig,
    u_int32_t cookie, struct uni_msg *m)
{
	printf("UNI sig %s\n", uni_signame(sig));
	user_output(arg, sig, cookie, m);
	if (m != NULL)
		uni_msg_destroy(m);
}

/*
 * Input from user
 */
static void
user_input(int fd, int eventmask __unused, void *uap)
{
	struct stk *stk = (struct stk *)uap;
	char line[LINE_MAX], *ptr, *end;
	ssize_t n;

	if ((n = read(fd, line, sizeof(line))) == -1)
		err(1, "read from user");
	if (n == 0) {
		Tcl_DStringFree(&command);
		Tcl_DStringAppend(&command, "exit", 4);
		try_exec(stk->interp, &command);
		exit(2);
	}
	ptr = line;
	while (n >= 0) {
		for (end = ptr; end < ptr + n && *end != '\n'; end++)
			;
		if (end == ptr + n) {
			Tcl_DStringAppend(&command, ptr, n);
			break;
		}
		Tcl_DStringAppend(&command, ptr, end - ptr);
		try_exec(stk->interp, &command);
		n -= end - ptr + 1;
		ptr = end + 1;
	}
}

static void
sscop_verbose(struct sscop *sscop __unused, void *arg __unused,
    const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	fprintf(stderr, "sscop: ");
	vfprintf(stderr, fmt, ap);
	fprintf(stderr, "\n");
	va_end(ap);
}

static void
sscfu_verbose(struct sscfu *sscfu __unused, void *arg __unused,
    const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	fprintf(stderr, "sscf: ");
	vfprintf(stderr, fmt, ap);
	fprintf(stderr, "\n");
	va_end(ap);
}

static void
uni_verbose(struct uni *uni __unused, void *arg __unused,
    enum uni_verb fac, const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	fprintf(stderr, "uni(%s): ", uni_facname(fac));
	vfprintf(stderr, fmt, ap);
	fprintf(stderr, "\n");
	va_end(ap);
}

static void
uni_status_func(struct uni *uni __unused, void *uarg __unused,
    void *arg __unused, const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
}


static const struct sscop_funcs sscop_funcs = {
	sscop_send_manage,
	sscop_send_upper,
	sscop_send_lower,
	sscop_verbose,
	sscop_start_timer,
	sscop_stop_timer
};

static const struct sscfu_funcs sscfu_funcs = {
	sscfu_send_upper,
	sscfu_send_lower,
	sscfu_window,
	sscfu_verbose
};

static const struct uni_funcs uni_funcs = {
	uni_uni_output,
	uni_saal_output,
	uni_verbose,
	uni_status_func,
	uni_start_timer,
	uni_stop_timer
};

int
main(int argc __unused, char *argv[] __unused)
{
	struct stk stk;
	Tcl_DString args;
	char *str;
	const char *res;
	int opt;

	memset(&stk, 0, sizeof(stk));
	stk.atm_fd = STDIN_FILENO;
	stk.user_fd = STDOUT_FILENO;

	argc--;
	argv++;

	if ((stk.sscop = sscop_create(&stk, &sscop_funcs)) == NULL)
		err(1, "cannot create sscop instance");
	if ((stk.sscf = sscfu_create(&stk, &sscfu_funcs)) == NULL)
		err(1, "cannot create sscf instance");
	if ((stk.uni = uni_create(&stk, &uni_funcs)) == NULL)
		err(1, "cannot create uni instance");

	if ((stk.interp = Tcl_CreateInterp()) == NULL)
		err(1, "cannot create TCL interpreter");
	Tcl_DStringInit(&command);

	user_tcl_init(&stk);

	/*
	 * Make script name available as 'script'
	 */
	if((str = malloc(strlen(script)+1)) == NULL)
		err(1, NULL);
	script = strcpy(str, script);
	if(Tcl_LinkVar(stk.interp, "script", (char *)&script, 
	    TCL_LINK_STRING|TCL_LINK_READ_ONLY) != TCL_OK)
		errx(1, "Tcl_LinkVar(script): %s",
		    Tcl_GetStringResult(stk.interp));

	/*
	 * Make the execution path of the script availablle as 'tclpath'
	 */
	if((str = malloc(strlen(tclpath)+1)) == NULL)
		err(1, NULL);
	tclpath = strcpy(str, tclpath);
	if(Tcl_LinkVar(stk.interp, "tclpath", (char *)&tclpath, 
	    TCL_LINK_STRING|TCL_LINK_READ_ONLY) != TCL_OK)
		errx(1, "Tcl_LinkVar(tclpath): %s",
		    Tcl_GetStringResult(stk.interp));

	/*
	 * Make the program arguments available
	 */
	Tcl_DStringInit(&args);
	while(argc--)
		Tcl_DStringAppendElement(&args, *argv++);
	Tcl_SetVar(stk.interp, "argv", Tcl_DStringValue(&args), 0);
	Tcl_DStringFree(&args);

	/*
	 * Run startup script
	 */
	asprintf(&str, "%s/%s", tclpath, script);
	if (str == NULL)
		err(1, "asprintf");
	opt = Tcl_EvalFile(stk.interp, str);
	free(str);

	if (*(res = Tcl_GetStringResult(stk.interp)) != 0)
		fprintf(stderr, "%s\n", res);
	if (opt != TCL_OK) {
		fprintf(stderr, "initialisation returned %d\n", opt);
		if (errinfo) {
			if ((res = Tcl_GetVar(stk.interp, "errorInfo",
			    TCL_GLOBAL_ONLY)) != NULL)
				fprintf(stderr, "errorInfo='%s'\n", res);
			if ((res = Tcl_GetVar(stk.interp, "errorCode",
			    TCL_GLOBAL_ONLY)) != NULL)
				fprintf(stderr, "errorCode='%s'\n", res);
		}
		exit(1);
	}

	/*
	 * Start file descriptors
	 */
	stk.atm_fdh = poll_register(stk.atm_fd, atm_input, &stk, POLL_IN);
	stk.user_fdh = poll_register(stk.user_fd, user_input, &stk, POLL_IN);

	doprompt = 1;
	while (1) {
		if (doprompt && stk.user_fdh != -1 && isatty(stk.user_fd)) {
			write(stk.user_fd, "@> ", 3);
			doprompt = 0;
		}
		poll_dispatch(1);
	}

	return (0);
}
